from .web.resource import Resource, root
from .web.shortcuts import page
from .conn import REDIS
from .execution import RECENT_KEY, utc_now

from datetime import timedelta, datetime

import pytz


utc = pytz.timezone('UTC')
local = pytz.timezone("America/New_York")


class MainDashboard(Resource):

    def render_GET(self, request):

        now = utc_now()
        past = now - timedelta(days=2)
        start = past.strftime('%s')
        finish = now.strftime('%s')
        recent = REDIS.zrevrangebyscore(
            RECENT_KEY, finish, start, withscores=True
        )
        out = []
        if recent:
            out = [
                {
                    'datetime': datetime.fromtimestamp(
                        timestamp).replace(tzinfo=utc).astimezone(local),
                    'taskname': task.split(':')[0],
                    'uid': task.split(':')[-1]
                } for task, timestamp in recent
            ]

        return page(request, 'main.html', {'data': out})


root.putChild('tiempo', MainDashboard())



from .. import PROJECT_PATH

from jinja2 import Environment, FileSystemLoader
import os

TEMPLATE_DIRS = [os.path.join(PROJECT_PATH, 'templates'), ]
JINJA_ENV = Environment(loader=FileSystemLoader(TEMPLATE_DIRS))



from twisted.application import service, internet
from twisted.internet import reactor
from twisted.python.threadpool import ThreadPool
from twisted.web.server import Site

import logging

logger = logging.getLogger(__name__)


class Service(service.MultiService):
    """
    """

    def __init__(self, resource, port=3333, **kwargs):
        service.MultiService.__init__(self)

        self.addThreadPool()
        self.addTCPServer('root', resource, port)

    def addThreadPool(self, **kwargs):
        # Create, start and add a thread pool service, which is made available
        # to our WSGIResource within HendrixResource
        threads = ThreadPool()
        reactor.addSystemEventTrigger('after', 'shutdown', threads.stop)
        thread_service = ThreadPoolService(threads)
        self.addNamedService('threads', thread_service)

    def addTCPServer(self, name, resource, port):
        factory = Site(resource)
        tcp_service = TCPServer(port, factory)
        self.addNamedService(name, tcp_service)

    def addNamedService(self, name, service):
        # to get this at runtime use service.getServiceNamed('service_name')
        service.setName(name)
        service.setServiceParent(self)

    def getPort(self, name):
        "Return the port object associated to our tcp server"
        service = self.getServiceNamed(name)
        return service._port


class ThreadPoolService(service.Service):
    '''
    A simple class that defines a threadpool on init
    and provides for starting and stopping it.
    '''
    def __init__(self, pool):
        "self.pool returns the twisted.python.ThreadPool() instance."
        if not isinstance(pool, ThreadPool):
            msg = '%s must be initialised with a ThreadPool instance'
            raise TypeError(
                msg % self.__class__.__name__
            )
        self.pool = pool

    def startService(self):
        service.Service.startService(self)
        self.pool.start()

    def stopService(self):
        service.Service.stopService(self)
        self.pool.stop()


class TCPServer(internet.TCPServer):

    def __init__(self, port, factory, *args, **kwargs):
        internet.TCPServer.__init__(self, port, factory, *args, **kwargs)
        self.factory = factory




import time
from tiempo.task import task


@task(periodic=True)
def test_task_logging():
    print 'test task logging is running'

    for i in range(1, 20):
        print 'doing nothing: %d' % i

    print 'spawned a task:', test_task_spawning.soon(
        True, veggies=['celery', 'carrot', 'beets']
    )
    time.sleep(3)
    print 'spawned a task:', test_task_spawning.soon(
        False, fruits=['banana', 'apple', 'cherry']
    )

    print 'sleeping for 60 secs'
    time.sleep(61)
    print 'slept for 60'


@task()
def test_task_spawning(*args, **kwargs):
    time.sleep(5)
    print 'this task ran just fine with args, kwargs: %r, %r' % (args, kwargs)




from datetime import datetime
from tiempo.execution import REDIS, RECENT_KEY
from tiempo import conf as tiemposettings

from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.http import HttpResponse

import pytz
import json
import dateutil.parser

utc = pytz.timezone('UTC')
local = pytz.timezone("America/New_York")




@login_required
def dashboard(request):

    queue_length = [
        {
        'name': t, 
        'length': REDIS.llen(t),
        'last': REDIS.get('last_finished_%s'%t),
        }
        for t in tiemposettings.TASK_GROUPS
    ]


    response = render(request, 'tiempo/dashboard.html', {
        'queue_info': queue_length
    })
    return response


@login_required
def recent_tasks(request):
    start = int(request.GET.get('offset', 0))
    end = start + int(request.GET.get('limit', 1000))

    recent = REDIS.zrevrange(RECENT_KEY, start, end, withscores=True)
    # recent.reverse()

    out = [
        {
            'datetime': datetime.fromtimestamp(
                timestamp).replace(tzinfo=utc).astimezone(local),
            'taskname': task.split(':')[0],
            'uid': task.split(':')[-1]
        } for task, timestamp in recent
    ]

    response = render(request, 'tiempo/recent.html', {
        'tasks': out,
    })
    return response


@login_required
def results(requests, key):

    # {u'task': u'apps.brand.tasks.update_youtube_member_channels_nightly',
    # u'uid': u'4856ab75-5964-11e4-9d8b-5cf938a858da',
    # u'start_time': u'2014-10-21T20:53:32.582777+00:00', u'end_time':
    # u'2014-10-21T20:53:32.602500+00:00', u'duration': 0.019723, u'output':
    # [u'update_youtube_member_channels_nightly',
    #  u'found 0 CampaignCreators needing updates']}
    task = json.loads(REDIS.get(key))
    content = {}

    for key, val in task.items():
        if 'time' in key:
            content[key] = dateutil.parser.parse(val).strftime('%a %H:%M:%S')
        elif key == 'output':
            content[key] = '<br>'.join(val)
        else:
            content[key] = val

    return HttpResponse(
        json.dumps(content),
        content_type='application/json',
        status=200
    )



from django.conf.urls import patterns, url

urlpatterns = patterns(
    'tiempo.contrib.django.views',
    url(r'^$', 'dashboard', name='tiempo_dashboard'),
    url(r'^recent/$', 'recent_tasks', name='recent_tasks'),
    url(r'^results/(?P<key>.+)', 'results', name='task_results')
)



from twisted.web import resource, static
from twisted.web._responses import NOT_FOUND, FORBIDDEN
from .shortcuts import page


class Resource(resource.Resource):

    resourceFactory = None

    def __init__(self, *args, **kwargs):
        resource.Resource.__init__(self)
        self.args = args
        self.kwargs = kwargs

    def render(self, request):
        return resource.Resource.render(self, request)

    def getChild(self, name, request):
        if not self.resourceFactory:
            return NotFound()
        return self.resourceFactory(name, request)


class ErrorPage(Resource):

    template = '400.html'

    def __init__(self, status, brief, detail):
        Resource.__init__(self)
        self.code = status
        self.brief = brief
        self.detail = detail

    def render(self, request):
        request.setResponseCode(self.code)
        request.setHeader(b"content-type", b"text/html; charset=utf-8")
        interpolated = page(
            request, self.template, {
                'code': self.code, 'brief': self.brief, 'detail': self.detail
            }
        )
        if isinstance(interpolated, unicode):
            return interpolated.encode('utf-8')
        return interpolated

    def getChild(self, name, request):
        return self


class NotFound(ErrorPage):

    template = '404.html'

    def __init__(self, message='Sorry, this page does not exist.'):
        ErrorPage.__init__(self, NOT_FOUND, 'Page Not Found', message)


class Forbidden(ErrorPage):

    template = '403.html'

    def __init__(self, message='Sorry, this listing is forbidden'):
        ErrorPage.__init__(self, FORBIDDEN, 'Forbidden', message)


class File(static.File):
    '''
    A simple static service with directory listing disabled
    Gives the client a 403 instead of letting them browse a static directory.
    '''

    # Resource used to render 404 Not Found error pages.
    childNotFound = NotFound()

    def directoryListing(self):
        # Override to forbid directory listing
        return Forbidden()


root = Resource()



from django.conf import settings
from tiempo.conf import DEBUG

import chalk
import importlib


def auto_load_tasks():
    for app in settings.PROJECT_APPS:
        module = importlib.import_module(app)
        try:
            importlib.import_module(app + '.tasks')
            chalk.blue('imported tasks from %s' % app)
        except ImportError as e:
            # print traceback.format_exc()
            pass
    if DEBUG:
        pass
        # import tasks




try:
    from django.utils.encoding import force_bytes
except ImportError:
    from .utils import force_bytes

try:
    from six.moves import cPickle as pickle
except ImportError:
    import pickle

from . import TIEMPO_REGISTRY
from .conn import REDIS

from logging import getLogger

import inspect
import uuid
import base64
import importlib
import functools
import datetime


logger = getLogger(__name__)


class TaskBase(object):

    @property
    def html(self):
        return '<task>%s</task>' % self.uid

    @classmethod
    def _encode(cls, dictionary):
        "Returns the given session dictionary pickled and encoded as a string."
        pickled = pickle.dumps(dictionary, pickle.HIGHEST_PROTOCOL)
        return base64.b64encode(pickled).decode('ascii')

    @staticmethod
    def _decode(data):

        encoded_data = base64.b64decode(force_bytes(data))
        try:
            # could produce ValueError if there is no ':'
            pickled = encoded_data
            dec = pickle.loads(pickled)
            # print 'pickle:', dec
            return dec

        except Exception:
            raise
            # ValueError, SuspiciousOperation, unpickling exceptions. If any of
            # these happen, just return an empty dictionary (an empty session).
            return {}

    @staticmethod
    def rehydrate(base_64):

        d = Task._decode(base_64)

        module = importlib.import_module(d['function_module_path'])
        T = getattr(module, d['function_name'])
        T.data = d
        T._thaw()
        return T


class Task(TaskBase):

    def __repr__(self):
        return self.key

    def __init__(self, *args, **kwargs):

        self.groups = ['ALL']

        self.day = None
        self.hour = None
        self.minute = None
        self.periodic = False

        self.uid = str(uuid.uuid1())

        self.key = self.uid
        self.function_name = None
        self.group = 'ALL'

        self.frozen = False
        # group and other attrs may be overridden here.
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __call__(self, *args, **kwargs):

        # we only want this to happen if this is being called
        # as a decorator, otherwise all "calls" are performed as
        # special functions ie. "now" or "soon" etc.
        if args and hasattr(args[0], '__call__'):
            self.func = args[0]
            self.cache = {}
            functools.update_wrapper(self, self.func)
            self.key = '%s.%s' % (
                inspect.getmodule(self.func).__name__, self.func.__name__
            )
            TIEMPO_REGISTRY[self.key] = self
            return self

        return self

    def _freeze(self, *args, **kwargs):

        self.data = {
            'function_module_path': inspect.getmodule(self.func).__name__,
            'function_name': self.func.__name__,
            'args_to_function': args,
            'kwargs_to_function': kwargs,
            'schedule': self.get_schedule(),
            'uid': str(uuid.uuid1()),
        }
        self.frozen = True

        return self.data

    def _thaw(self, data=None):
        if not data and hasattr(self, 'data'):
            data = self.data

        if data:
            for key, val in data.items():
                setattr(self, key, val)

    def _get_function(self):
        if hasattr(self, 'func'):
            return self.func

        if not self.function_name:
            self._thaw()

        module = importlib.import_module(self.function_module_path)
        obj = getattr(module, self.function_name)

        if hasattr(obj, 'func'):
            return obj.func
        return obj

    def _enqueue(self):
        if not self.frozen:
            raise ValueError(
                'need to freeze this task before enqueuing'
            )

        d = self._encode(self.data)
        REDIS.rpush(self.group, d)
        return self.data['uid']

    def get_schedule(self):
        if self.periodic:
            sched = [
                '*',
                '*',
                '*'
            ]
            for i, inc in enumerate(['day', 'hour', 'minute']):
                attr = getattr(self, inc, None)
                if attr is not None:
                    if attr != '*':
                        attr = '%02d' % int(attr)
                    sched[i] = attr

            return '.'.join(sched)

    def run(self):
        """
            runs this task right now
        """
        self.start_time = datetime.datetime.now()
        func = self._get_function()
        func(
            *getattr(self, 'args_to_function', ()),
            **getattr(self, 'kwargs_to_function', {})
        )

    def soon(self, *args, **kwargs):
        """
            schedules this task to be run with the args and kwargs
            whenever a worker participating in this task's groups
            comes up as available

        """
        logger.debug('scheduling task %r for next available execution', self)
        self._freeze(*args, **kwargs)
        self._enqueue()
        return self

    def now(self, *args, **kwargs):
        """
            runs this task NOW with the args and kwargs
        """
        self._freeze(*args, **kwargs)
        self.run()
        return self

    def finished(self, error=None):
        finished_text = """
%(key)s:
started at %(start)s
finished in %(elapsed)s seconds
%(errors)s"""%{
            'key':self.key,
            'start':self.start_time.strftime('%y/%m/%d %I:%M%p'),
            'elapsed':(datetime.datetime.now()-self.start_time).seconds,
            'errors': 'with errors: %s'%error if error else ''
        }

        REDIS.set('last_finished_%s'%self.group, finished_text)




task = Task
